#!/usr/bin/env python
#----------------------------------------------------------------------------
# Name:         PyTAS.py
# Purpose:      A Python based version of ISU TAS
#
# Author:       Matthew Patitz
#
# Created:      
# RCS-ID:       
# Copyright:    
# License:      
#----------------------------------------------------------------------------

"""
This simulator is designed to simulate aTAM systems based on the ISU TAS format

Requirements:
    This application requires Python 2.x due to its dependency on wxPython

Dependencies:
    wxPython    (download installer for your platform here: https://wxpython.org/Phoenix/snapshot-builds/
                I used this command:  "pip install -U --pre -f https://wxpython.org/Phoenix/snapshot-builds/ wxPython"
                for Windows and Mac (sudo -H pip install wxPython), and additional instructions can be found here:
                https://wxpython.org/Phoenix/snapshot-builds/README.txt)
    pil         (via pillow, install using 'pip install pillow', pip can be found in your Python27\Scripts dir)
    svgwrite    (via 'pip install svgwrite')

TODO:
    - implement tile type window to more clearly show tile type labels
    - create cached set of bitmaps for tile types (with and without labels)
    - implement clickable assembly overview window
    - interface for TAScript
    - tile type editor
    - packages for Windows and OS X

"""

#
#Previous dependencies no longer used:
#    pyPdf2      (install with 'pip install pypdf2')
#    reportlab   (install with 'pip install reportlab')


#import wxversion
#wxversion.select('3.0-msw', optionsRequired=True)
#wxversion.select('3.0-Phoenix', optionsRequired=True)
import wx
import wx.richtext as rt
from PyTASAssembly import *
from PyTASCanvas import *
from SimulatorFramework.SimApp import *
from SimulatorFramework import SimUtils
from PyTAS3DCanvas import *
from wx.lib.pubsub import pub
#from TileType import *
import wx.lib.agw.cubecolourdialog as CCD
import wx.aui
import wx.lib.intctrl
#from wx.lib.pubsub import Publisher


#===============================================================================
# from reportlab.lib.pagesizes import letter
# from reportlab.lib.units import inch
# from reportlab.pdfgen import canvas
# from reportlab.platypus import SimpleDocTemplate, Image
#===============================================================================

#---------------------------------------------------------------------------

class PyTASApp(SimApp):
    def __init__(self):
        self.dimMode = SIMU_2D_MODE
        self.tilesPerStepPanel = None
        SimApp.__init__(self)
        self.frame.Bind(wx.EVT_CHAR, self.OnChar)

    def OnInit(self):
        return SimApp.OnInit(self, title="PyTAS aTAM Simulator")

    def CreateCanvas(self):
        if self.dimMode == SIMU_2D_MODE:
            self.canvas = PyTASCanvas(self.frame, wx.ID_ANY, self.frame, wx.DefaultSize, self)
        elif self.dimMode == SIMU_3D_MODE:
            self.canvas = PyTAS3DCanvas(self.frame, wx.ID_ANY, self.frame, wx.DefaultSize, self)
        
    def CreateMenu(self):
        menuBar = wx.MenuBar()
        menu = wx.Menu()
        item1 = menu.Append(-1, "&Load\tCtrl-L", "Load system")
        self.Bind(wx.EVT_MENU, self.OnLoad, item1)
        itemR = menu.Append(-1, "Reload files", "Reload system from files")
        self.Bind(wx.EVT_MENU, self.OnReload, itemR)
        itemReset = menu.Append(-1, "&Reset to seed\tCtrl-R", "Reset system to seed")
        self.Bind(wx.EVT_MENU, self.OnReset, itemReset)
        item2 = menu.Append(-1, "&Save\tCtrl-S", "Save system")
        self.Bind(wx.EVT_MENU, self.OnSave, item2)
        menu.AppendSeparator()
        item3 = menu.Append(-1, "E&xit\tCtrl-X", "Exit simulator")
        self.Bind(wx.EVT_MENU, self.OnExitApp, item3)
        menuBar.Append(menu, "&File")

        menuCtrl = wx.Menu()
        
        self.radio2DMode = wx.MenuItem(menuCtrl, -1, text = "2D mode", kind = wx.ITEM_RADIO) 
        self.radio3DMode = wx.MenuItem(menuCtrl, -1, text = "3D mode", kind = wx.ITEM_RADIO)
        menuCtrl.Append(self.radio2DMode)
        menuCtrl.Append(self.radio3DMode)
        self.radio2DMode.Check()
        self.Bind(wx.EVT_MENU, self.OnDimMode, self.radio2DMode)
        self.Bind(wx.EVT_MENU, self.OnDimMode, self.radio3DMode)
        menuCtrl.AppendSeparator()
        
        self.radioIgnoreNondet = wx.MenuItem(menuCtrl, -1, text = "Ignore nondeterminism", kind = wx.ITEM_RADIO) 
        self.radioReportNondet = wx.MenuItem(menuCtrl, -1, text = "Report nondeterminism", kind = wx.ITEM_RADIO)
        self.radioStopAtNondet = wx.MenuItem(menuCtrl, -1, text = "Halt on nondeterminism", kind = wx.ITEM_RADIO)
        menuCtrl.Append(self.radioIgnoreNondet)
        menuCtrl.Append(self.radioReportNondet)
        menuCtrl.Append(self.radioStopAtNondet)
        self.radioReportNondet.Check()
        self.Bind(wx.EVT_MENU, self.OnNondetMode, self.radioIgnoreNondet)
        self.Bind(wx.EVT_MENU, self.OnNondetMode, self.radioReportNondet)
        self.Bind(wx.EVT_MENU, self.OnNondetMode, self.radioStopAtNondet)
        menuCtrl.AppendSeparator()
        self.checkKeepCache = wx.MenuItem(menuCtrl, -1, text = "Keep cache for rewinding", kind = wx.ITEM_CHECK)
        menuCtrl.Append(self.checkKeepCache)
        self.checkKeepCache.Check(True)
        self.Bind(wx.EVT_MENU, self.OnKeepCache, self.checkKeepCache)
        menuCtrl.AppendSeparator()
        self.setTilesPerStep = wx.MenuItem(menuCtrl, -1, text = "Set tile additions per step...")
        menuCtrl.Append(self.setTilesPerStep)
        self.Bind(wx.EVT_MENU, self.OnSetTilesPerStep, self.setTilesPerStep)
        menuCtrl.AppendSeparator()
        itemFR = menuCtrl.Append(-1, "Fast rewind", "Fast rewind")
        self.Bind(wx.EVT_MENU, self.OnFastRewind, itemFR)
        itemStepRew = menuCtrl.Append(-1, "Rewind", "Rewind one step")
        self.Bind(wx.EVT_MENU, self.OnRewind, itemStepRew)
        itemStop = menuCtrl.Append(-1, "Pause", "Stop simulation")
        self.Bind(wx.EVT_MENU, self.OnQuit, itemStop)
        itemStep = menuCtrl.Append(-1, "&Forward\tCtrl-F", "Forward one step")
        self.Bind(wx.EVT_MENU, self.OnForward, itemStep)
        itemFF = menuCtrl.Append(-1, "&Go\tCtrl-G", "Fast forward")
        self.Bind(wx.EVT_MENU, self.OnGo, itemFF)
        itemQuit = menuCtrl.Append(-1, "&Quit\tCtrl-Q", "Quit fast forward")
        self.Bind(wx.EVT_MENU, self.OnQuit, itemQuit)
        menuBar.Append(menuCtrl, "&Control")

        self.menuView = wx.Menu()
        itemSTC = self.menuView.Append(-1, "&Center\tCtrl+C", "Scroll to center")
        self.Bind(wx.EVT_MENU, self.OnCenter, itemSTC)
        itemCOC = self.menuView.Append(-1, "&Center on coordinates...", "Scroll coordinates to center")
        self.Bind(wx.EVT_MENU, self.OnCenterOnCoords, itemCOC)
        self.menuView.AppendSeparator()
        itemZIn = self.menuView.Append(-1, "Zoom in\tCtrl-+", "Zoom in")
        self.Bind(wx.EVT_MENU, self.OnZoomIn, itemZIn)
        itemZOut = self.menuView.Append(-1, "Zoom out\tCtrl--", "Zoom out")
        self.Bind(wx.EVT_MENU, self.OnZoomOut, itemZOut)
        self.menuView.AppendSeparator()
        itemTG = self.menuView.Append(-1, "&Toggle grid lines\tCtrl-T", "Toggle display of grid lines")
        self.Bind(wx.EVT_MENU, self.OnToggleGrid, itemTG)
        itemTD = self.menuView.AppendCheckItem(-1, "&Toggle display updates\tCtrl-D", "Toggle display of tile additions")
        self.Bind(wx.EVT_MENU, self.OnToggleDisplay, itemTD)
        itemSTT = self.menuView.Append(-1, "Show only matching tile types...", "Set patterns for tile type names to display")
        self.Bind(wx.EVT_MENU, self.OnShowMatching, itemSTT)
        self.menuView.AppendSeparator()
        itemBGC = self.menuView.Append(-1, "Set background color", "Set background color")
        self.Bind(wx.EVT_MENU, self.OnSetBackgroundColor, itemBGC)
        self.menuView.AppendSeparator()
        itemRefresh = self.menuView.Append(-1, "Refresh screen\tF2", "Refresh screen")
        self.Bind(wx.EVT_MENU, self.OnRefresh, itemRefresh)

        menuBar.Append(self.menuView, "&View")

        menu2 = wx.Menu()
        itemAbout = menu2.Append(-1, "About", "About")
        self.Bind(wx.EVT_MENU, self.OnAbout, itemAbout)
        itemH = menu2.Append(-1, "&Help\tCtrl-H", "Show help")
        self.Bind(wx.EVT_MENU, self.OnHelp, itemH)
        menuBar.Append(menu2, "&Help")
        
        self.frame.SetMenuBar(menuBar)
                
    def OnLoad(self, evt):
        openFileDialog = wx.FileDialog(self.canvas, "Open", "", "", 
                                       "Tile system files (*.tdp)|*.tdp|TAMScript python files (*.py)|*.py",
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):
            
            dlg = wx.TextEntryDialog(self.canvas, "You can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.canvas.randSeed))
            if dlg.ShowModal() == wx.ID_OK:
                self.canvas.randSeed = int(dlg.GetValue())
                dlg.Destroy()
            else:
                dlg.Destroy()
                return False

            self.canvas.SetRandomSeed(self.canvas.randSeed)
            self.OutputMessage("Loading tile assembly system...")
            #self.InitializeDrawing()
            if self.dimMode == SIMU_3D_MODE:
                self.canvas.assembly.seedPatch = []
                self.canvas.Reset(bFromFiles=False, bForce=True)
                self.canvas.assembly = PyTAS3DAssembly()
            else:
                self.canvas.assembly = PyTASAssembly()
                
            ## Determine .tds or .py
            moduleName,fileExtension = os.path.splitext(os.path.split(fileName)[-1])
            self.canvas.assembly.useScript = (fileExtension.lower() == '.py')
            if self.canvas.assembly.useScript: ## script
                print("Using script")
            else:                              ## .tds/.tdp format
                print("Not using script")
                
            midPt = self.canvas.assembly.LoadTileSystem(fileName, backgroundColor=self.canvas.backgroundColor)
            
            if midPt == TILESET_LOAD_FAIL_NEEDS_3D:
                dlg = wx.MessageDialog(self.frame, "The selected system requires 3D mode, automatically switching.", caption="Switching to 3D mode")
                if dlg.ShowModal() == wx.ID_OK:
                    self.SetDimMode(SIMU_3D_MODE)
                    #self.canvas.assembly.temperature = savedAssembly.temperature
                    #self.canvas.assembly.tileTypes = savedAssembly.tileTypes
                    #self.canvas.assembly.seedPatch = savedAssembly.seedPatch
                    self.canvas.Reset(bFromFiles=False, bForce=True)
                    midPt = self.canvas.assembly.LoadTileSystem(fileName, backgroundColor=self.canvas.backgroundColor)
                    self.canvas.RedrawAssembly()
                else:
                    dlg.Destroy()
            elif (self.canvas.assembly.bHas3D == False) and (self.dimMode == SIMU_3D_MODE):
                dlg = wx.MessageDialog(self.frame, "The selected system does not require 3D mode.\nWould you like to switch to 2D?", caption="Switch to 2D mode?", style=wx.YES_NO)
                if dlg.ShowModal() == wx.ID_YES:
                    print("TODO: Implement the auto switch to 2D...")
                    dlg.Destroy()
            self.canvas.RedrawAssembly(checkAll=True)
            self.canvas.ScrollToCoords(midPt)
            self.OutputMessage("Tile assembly system loaded")
            return True

    def OnToggleDisplay(self, event):
        self.canvas.drawUpdates = (self.canvas.drawUpdates + 1) % 2
        if self.canvas.drawUpdates == SimCanvas.SIMFW_FULLDISPLAY:
            if len(self.canvas.displayPatterns) == 0:
                self.OutputMessage("Switched display mode:  Displaying all tiles in the assembly")
            else:
                self.OutputMessage("Switched display mode:  Displaying all tiles matching the following patterns: " + str(self.canvas.displayPatterns))
            if self.dimMode == SIMU_2D_MODE:
                self.canvas.UpdateSizeVars()
                #self.RedrawAssembly()
                if not self.canvas.savedCenteredCoords is None:
                    self.canvas.CanvasSizeUpdate(self.canvas.savedCenteredCoords)
            elif self.dimMode == SIMU_3D_MODE:
                self.canvas.Refresh()
            self.canvas.SendStatusBarUpdate(1, "Number of tiles: " + str(self.canvas.assembly.numTiles))
        #elif self.drawUpdates == SIMFW_GROWTHDISPLAY:
        #    SimUtils.globalApp.OutputMessage("Switched display mode:  Displaying only addition of growth tiles")
        elif self.canvas.drawUpdates == SimCanvas.SIMFW_NODISPLAY:
            if self.dimMode == SIMU_2D_MODE:
                self.canvas.savedCenteredCoords = self.canvas.GetCenterCoords()
            self.OutputMessage("Switched display mode:  No tile additions are being displayed")

    def OnDimMode(self, event):
        if self.radio2DMode.IsChecked():
            self.SetDimMode(SIMU_2D_MODE)
        elif self.radio3DMode.IsChecked():
            self.SetDimMode(SIMU_3D_MODE)
        
    def SetDimMode(self, mode):
        if mode == self.dimMode:    # No change
            return
        if (not mode == SIMU_2D_MODE) and (not mode == SIMU_3D_MODE):
            SimUtils.globalApp.OutputMessage("Invalid dimension mode requested: " + str(mode))
            return

        #NOTE: creation of a new canvas re-seeds the random number generator, which may not be desired
        self.dimMode = mode
        
        if self.dimMode == SIMU_2D_MODE:
            self.radio2DMode.Check(True)
            self.SetCaching(True)
            self.checkKeepCache.Enable(True)
        else:
            self.radio3DMode.Check(True)
            self.SetCaching(False)
            self.checkKeepCache.Enable(False)

        pane = self.frame._mgr.GetPane(self.canvas) 
        if pane.IsOk():
            self.canvas.UnsubscribeEventHandlers()
            #self.frame._mgr.ClosePane(pane) 
            self.frame._mgr.DetachPane(pane.window) 
            pane.window.Destroy() 

        self.CreateCanvas()
        self.frame._mgr.AddPane(self.canvas, wx.CENTER, 'Simulation Window')
        self.canvas.SubscribeEventHandlers()
        
        if self.dimMode == SIMU_3D_MODE:
            self.CreateTilesPerStepWin()
        #else:
        #    self.frame._mgr.DetachPane()

        self.frame._mgr.Update()
        
    def CreateStatusBar(self):
        self.frame.CreateStatusBar(3)
        self.frame.SetStatusWidths([140,250,-1])
        
    def CreateTilesPerStepWin(self):
        if self.tilesPerStepPanel is None:
            self.tilesPerStepPanel = wx.Panel(self.frame, -1, wx.DefaultPosition, wx.DefaultSize)
            self.tilesPerStepLabel = wx.StaticText(self.tilesPerStepPanel, -1, "Tiles Per Step", wx.DefaultPosition, wx.Size(50,15))
            self.numTilesPerStepCtrl = wx.lib.intctrl.IntCtrl(self.tilesPerStepPanel, PYTAS_ID_STEPCOUNT, 1, wx.DefaultPosition, wx.Size(50,25))
            self.numTilesPerStepCtrl.SetMin(1)
            self.numTilesPerStepCtrl.SetNoneAllowed(False)
            #self.numTilesPerStepCtrl.SetLimited(True)
            self.numTilesPerStepCtrl.Bind(wx.EVT_ENTER_WINDOW, self.OnMouseEnter)
            self.numTilesPerStepCtrl.Bind(wx.EVT_LEAVE_WINDOW, self.OnMouseLeave)
            self.fullFrontierCheck = wx.CheckBox(self.tilesPerStepPanel, -1, "Full Frontier", wx.DefaultPosition, wx.DefaultSize)
            self.fullFrontierCheck.Bind(wx.EVT_LEAVE_WINDOW, self.OnMouseLeave)
            self.fullFrontierCheck.Bind(wx.EVT_CHECKBOX, self.OnFullFrontierCheck)
            
            #self.numTilesPerStepCtrl.Bind(wx.EVT_CHAR, self.OnTextCtrlChar)
            tsz = wx.BoxSizer(wx.HORIZONTAL)
            tsz.AddSpacer(10)
            tsz.Add(self.tilesPerStepLabel, 1, wx.ALIGN_CENTER_VERTICAL)
            tsz.Add(self.numTilesPerStepCtrl, 1, wx.ALIGN_CENTER_VERTICAL)
            tsz.AddSpacer(10)
            tsz.Add(self.fullFrontierCheck, 1, wx.ALIGN_CENTER_VERTICAL)
            tsz.AddSpacer(10)
            self.tilesPerStepPanel.SetSizerAndFit(tsz)
            self.frame._mgr.AddPane(self.tilesPerStepPanel, wx.aui.AuiPaneInfo().Name("Set Tiles Per Step").ToolbarPane().Top().Row(1).Hide())
            self.frame._mgr.Update()
            
    def OnShowTilesPerStepWin(self, event):
        self.CreateTilesPerStepWin()
        pane = self.frame._mgr.GetPane(self.tilesPerStepPanel)
        if pane.IsShown() == False:
        #if self.itemShowMess.IsChecked():
            pane.Show()
        else:
            pane.Hide()
        self.frame._mgr.Update()

    def OnSetTilesPerStep(self, evt):
        self.OnShowTilesPerStepWin(evt)
        #self.numTilesPerStepCtrl.SetFocus()
        
    def SetNumTilesPerStepEnabled(self, b):
        if self.canvas.stepMode == PYTAS_SM_FULL:
            b = False
        if b:
            self.numTilesPerStepCtrl.SetEditable(True)
            self.numTilesPerStepCtrl.SetBackgroundColour((255, 255, 255))
        else:
            self.numTilesPerStepCtrl.SetEditable(False)
            self.numTilesPerStepCtrl.SetBackgroundColour((200, 200, 200))
            
    def OnFullFrontierCheck(self, evt):
        if self.fullFrontierCheck.IsChecked():
            self.canvas.stepMode = PYTAS_SM_FULL
            self.SetNumTilesPerStepEnabled(False)
            self.OutputMessage("Entering full frontier mode for tile additions per step")
        else:
            self.canvas.stepMode = PYTAS_SM_STEP
            self.SetNumTilesPerStepEnabled(True)
            self.OutputMessage("Leaving full frontier mode for tile additions per step")
        self.frame.SetFocus()

    def OnMouseEnter(self, evt):
        evt.Skip()
        
    def OnMouseLeave(self, evt):
        self.canvas.numTilesPerStep = self.numTilesPerStepCtrl.GetValue()
        if self.canvas.numTilesPerStep < 1:
            self.canvas.numTilesPerStep = 1
        self.frame.SetFocus()
    
    def OnCenterOnCoords(self, event):
        if self.dimMode == SIMU_3D_MODE:
            doAgain = True
            coords = (0,0,0)
            lookat = (0,0,1)
            up = (0,1,0)
            while doAgain == True:
                doAgain = False
                dlg = wx.TextEntryDialog(self.frame, "Enter the coordinates to center on the camera\nin the form (x,y,z):", caption="Enter coordinates", value=str(coords))
                if not dlg.ShowModal() == wx.ID_OK:
                    return
                result = dlg.GetValue()
                dlg.Destroy()
                try:
                    c = eval(result)
                    coords = (c[0],c[1],c[2])
                except:
                    doAgain = True

            doAgain = True

            while doAgain == True:
                doAgain = False
                dlg = wx.TextEntryDialog(self.frame, "Enter a non-zero direction vector to point the camera\nin the form (x,y,z):", caption="Enter direction", value=str(lookat))
                if not dlg.ShowModal() == wx.ID_OK:
                    return
                result = dlg.GetValue()
                dlg.Destroy()
                try:
                    c = eval(result)
                    lookat = (c[0],c[1],c[2])
                    if lookat == (0,0,0):
                        doAgain = True
                except:
                    doAgain = True

            doAgain = True

            while doAgain == True:
                doAgain = False
                dlg = wx.TextEntryDialog(self.frame, "Enter a non-zero direction vector which points up from the camera\nin the form (x,y,z):", caption="Enter direction", value=str(up))
                if not dlg.ShowModal() == wx.ID_OK:
                    return
                result = dlg.GetValue()
                dlg.Destroy()
                try:
                    c = eval(result)
                    up = (c[0],c[1],c[2])
                    if up == (0,0,0):
                        doAgain = True
                except:
                    doAgain = True

            self.canvas.CameraLookAt(coords, ([coords[i] + lookat[i] for i in range(3)]), up)
        elif self.dimMode == SIMU_2D_MODE:
            doAgain = True
            coords = (0,0)
            while doAgain == True:
                doAgain = False
                dlg = wx.TextEntryDialog(self.frame, "Enter the coordinates to center on the screen\nin the form (x,y):", caption="Enter coordinates", value=str(coords))
                if not dlg.ShowModal() == wx.ID_OK:
                    return
                result = dlg.GetValue()
                dlg.Destroy()
                try:
                    c = eval(result)
                    coords = (c[0],c[1])
                except:
                    doAgain = True
            self.canvas.ScrollToCoords(coords)
        else:
            return

    #===========================================================================
    # def OnTextCtrlChar(self, evt):
    #     print("getting keycode")
    #     keycode = evt.GetKeyCode()
    #     if keycode == wx.WXK_RETURN:
    #         print("got here")
    #         self.frame.SetFocus()
    #     else:
    #         evt.Skip()
    #         print(str(self.numTilesPerStepCtrl.GetValue()))
    #===========================================================================

#----------------------------------------------------------------------------

app = PyTASApp()
app.MainLoop()

#---------------------------------------------------------------------------
