import wx
from SimulatorFramework import *
from SimulatorFramework.TileType import *


N = North = 0
E = East = 1
S = South = 2
W = West = 3
U = Up = 4
D = Down = 5

DirAbbrevArray = ("N", "E", "S", "W", "U", "D")
DirNameArray = ("NORTH", "EAST", "SOUTH", "WEST", "UP", "DOWN")

def GetFontSizeToFitTextInBounds(text, font, dc, width, height):
    retFont = font
    (w, pixelSize) = retFont.GetPixelSize()
    bContinue = True
    while bContinue == True:
        dc.SetFont(retFont)
        (w, h) = dc.GetTextExtent(text)
        if ((w <= width) and (h <= height)) or pixelSize <= 2:
            bContinue = False
        else:
            pixelSize = pixelSize - 2
            retFont.SetPixelSize((0, pixelSize))
    #(w, pixelSize) = retFont.GetPixelSize()
    #print("pixelSize: " + str(pixelSize))
    return retFont


CELLSIZE = 60
CELLSIZE = 60

# for 3D simulation
#----------------------------------------------------------
# in OpenGL 1.0 texture sides should be 2^n
TEXTURE_RES = 1024

# size of a side of a tile on the texture
SIDE_RES = int(TEXTURE_RES / 3)

# texture offsets of given side [N,E,S,W,U,D]
SIDE_OFFSET = [(0,0),(0,SIDE_RES),(SIDE_RES,0),(SIDE_RES,SIDE_RES),(2*SIDE_RES,0),(2*SIDE_RES,SIDE_RES)]
#----------------------------------------------------------

class PyTASTileType(TileType):
    """A representation of a tile type for use in the ISU TAS simulator.
       Every tile must have a unique name and a unique set of 4 glues. If
       a name is not explicitly provided, one will be formed automatically
       as a concatenation of the glues values of each side."""

    name = ""
    label = ""
    color = "White"
    colorVector = (255, 255, 255)
    sides = {}
    dims  = 2
    img   = None

    def __init__(self, name="", width=CELLSIZE, height=CELLSIZE, is3D=False):
        TileType.__init__(self, name, width, height)
        if is3D == False:
            self.dims = 2
            self.sides = {N:Glue(),E:Glue(),S:Glue(),W:Glue()}
        else:
            self.dims = 3
            self.sides = {N:Glue(),E:Glue(),S:Glue(),W:Glue(),U:Glue(),D:Glue()}
           
    def SetDimensions(self, dims):
        if self.dims == dims:
            return
        elif dims == 3:
            self.dims = 3
            self.sides[U] = Glue()
            self.sides[D] = Glue()
        elif dims == 2:
            self.dims = 2
            savedSides = self.sides
            self.sides = {N:savedSides[N], E:savedSides[E], S:savedSides[S], W:savedSides[W]}
        else:
            print("ERROR: invalid number of tile type dimensions: " + str(dims))

    def DrawAtLocation(self, xPixel, yPixel, pdc):
        if self.img is None:
            self.MakeImage()
        pdc.DrawBitmap(self.img, xPixel, yPixel)

    def MakeImage(self, backgroundColor=None):
            w = CELLSIZE*3
            h = CELLSIZE*3
            bmp = wx.Bitmap(w, h)
            dc = wx.MemoryDC()
            dc.SelectObject(bmp)
            
            if not backgroundColor is None:
                bg = wx.Brush(backgroundColor)#(self.GetBackgroundColour())
                dc.SetBackground(bg)
            dc.Clear()

            color = self.colorVector
            brush = wx.Brush(color)
            dc.SetBrush(brush)
            
            tabSize = 10
            dc.DrawRectangle(0+tabSize, 0+tabSize, w-(2*tabSize), h-(2*tabSize))
            
            color = wx.TheColourDatabase.Find("Black")
            brush = wx.Brush(color)
            dc.SetBrush(brush)

            # Draw northern tabs
            numTabs = self.sides[N].strength
            if numTabs > 0:
                spacing = w / (numTabs+1)
                currLoc = spacing
                for n in range(numTabs):
                    dc.DrawRectangle(currLoc-(tabSize/2), 0, tabSize, tabSize)
                    currLoc = currLoc + spacing

            # Draw southern tabs
            numTabs = self.sides[S].strength
            if numTabs > 0:
                spacing = w / (numTabs+1)
                currLoc = spacing
                for n in range(numTabs):
                    dc.DrawRectangle(currLoc-(tabSize/2), h-tabSize, tabSize, tabSize)
                    currLoc = currLoc + spacing
            
            # Draw western tabs
            numTabs = self.sides[W].strength
            if numTabs > 0:
                spacing = h / (numTabs+1)
                currLoc = spacing
                for n in range(numTabs):
                    dc.DrawRectangle(0, currLoc-(tabSize/2), tabSize, tabSize)
                    currLoc = currLoc + spacing

            # Draw eastern tabs
            numTabs = self.sides[E].strength
            if numTabs > 0:
                spacing = h / (numTabs+1)
                currLoc = spacing
                for n in range(numTabs):
                    dc.DrawRectangle(w-tabSize, currLoc-(tabSize/2), tabSize, tabSize)
                    currLoc = currLoc + spacing

            # Draw tile label
            wLabel = (w - 2*tabSize)/2
            hLabel = (h - 2*tabSize)/2
            defPixHt = 100
            ftype = wx.FONTFAMILY_DEFAULT#wx.FONTFAMILY_MODERN
            glueLabelScale = 0.6
            
            fname = ""#"Arial Narrow"
            #font = wx.Font(pointSize = 30, family = wx.DEFAULT, style = wx.NORMAL, weight = wx.NORMAL, faceName = 'Arial')
            if not self.label is None:
                font = wx.Font(pixelSize = (0,defPixHt), family = ftype, style = wx.NORMAL, weight = wx.FONTWEIGHT_BOLD, faceName = fname)#, faceName = 'Arial')
                font = GetFontSizeToFitTextInBounds(self.label, font, dc, wLabel, hLabel)
                (wLabel, hLabel) = dc.GetTextExtent(self.label)
                dc.DrawText(self.label, (w/2)-(wLabel/2), (h/2)-(hLabel/2))
            
            # Draw north glue label
            wLabel = ((w - 2*tabSize)/2) * glueLabelScale
            hLabel = ((h - 2*tabSize)/2) * glueLabelScale
            font = wx.Font(pixelSize = (0,defPixHt), family = ftype, style = wx.NORMAL, weight = wx.FONTWEIGHT_NORMAL, faceName = fname)
            font = GetFontSizeToFitTextInBounds(self.sides[N].label, font, dc, wLabel, hLabel)
            (wLabel, hLabel) = dc.GetTextExtent(self.sides[N].label)
            dc.DrawText(self.sides[N].label, (w/2)-(wLabel/2), (h/4)-(hLabel/2))
            
            # Draw south glue label
            wLabel = ((w - 2*tabSize)/2) * glueLabelScale
            hLabel = ((h - 2*tabSize)/2) * glueLabelScale
            font = wx.Font(pixelSize = (0,defPixHt), family = ftype, style = wx.NORMAL, weight = wx.FONTWEIGHT_NORMAL, faceName = fname)
            font = GetFontSizeToFitTextInBounds(self.sides[S].label, font, dc, wLabel, hLabel)
            (wLabel, hLabel) = dc.GetTextExtent(self.sides[S].label)
            dc.DrawText(self.sides[S].label, (w/2)-(wLabel/2), (3*h/4)-(hLabel/2))
            
            # Draw west glue label
            wLabel = ((w - 2*tabSize)/2) * glueLabelScale
            hLabel = ((h - 2*tabSize)/2) * glueLabelScale
            font = wx.Font(pixelSize = (0,defPixHt), family = ftype, style = wx.NORMAL, weight = wx.FONTWEIGHT_NORMAL, faceName = fname)
            font = GetFontSizeToFitTextInBounds(self.sides[W].label, font, dc, wLabel, hLabel)
            (wLabel, hLabel) = dc.GetTextExtent(self.sides[W].label)
            dc.DrawRotatedText(self.sides[W].label, (w/4)-1.5*(hLabel/2), (h/2)+(wLabel/2), 90.0)

            # Draw east glue label
            wLabel = ((w - 2*tabSize)/2) * glueLabelScale
            hLabel = ((h - 2*tabSize)/2) * glueLabelScale
            font = wx.Font(pixelSize = (0,defPixHt), family = ftype, style = wx.NORMAL, weight = wx.FONTWEIGHT_NORMAL, faceName = fname)
            font = GetFontSizeToFitTextInBounds(self.sides[E].label, font, dc, wLabel, hLabel)
            (wLabel, hLabel) = dc.GetTextExtent(self.sides[E].label)
            dc.DrawRotatedText(self.sides[E].label, (3*w/4)+1.5*(hLabel/2), (h/2)-(wLabel/2), 270.0)

            dc.SelectObject(wx.NullBitmap)
            
            img = bmp.ConvertToImage()
            img = img.Scale(CELLSIZE, CELLSIZE, wx.IMAGE_QUALITY_HIGH)
            bmp = wx.Bitmap(img)
            self.img = bmp

    def Copy(self):
        if self.dims == 2:
            t = TileType()
        else:
            t = TileType(True)
        t.sides[N] = self.sides[N].Copy()
        t.sides[E] = self.sides[E].Copy()
        t.sides[S] = self.sides[S].Copy()
        t.sides[W] = self.sides[W].Copy()
        if self.dims == 3:
            t.sides[U] = self.sides[U].Copy()
            t.sides[D] = self.sides[D].Copy()
        t.name = self.name
        t.label = self.label
        t.color = self.color
        t.colorVector = self.colorVector
        return t

    def SetInputSides(self, inputSet):
        for k in inputSet:
            self.sides[k] = Glue(inputSet[k].strength, inputSet[k].label)


    # function to parse the color string into a color vector
    def ParseColor(self):
        col = wx.TheColourDatabase.Find(self.color)
        if col.IsOk():
            self.colorVector = (col.Red(), col.Green(), col.Blue())
        else:
            if not self.color.lower().startswith("rgb"):
                self.colorVector = (255, 255, 255)
            else:
                channels = self.color[4:-1].split(",")
                try:
                    r = int(channels[0].strip())
                    g = int(channels[1].strip())
                    b = int(channels[2].strip())
                    self.colorVector = (r, g, b)
                except:
                    #print "Color Error on Tile: {}".format(self.name)#DEBUG
                    self.colorVector = (255, 255, 255)

