#!/usr/bin/env python
#----------------------------------------------------------------------------
# Name:         SimApp.py
# Purpose:      A basic, somewhat generic simulator application
#
# Author:       Matthew Patitz
#
# Created:      
# RCS-ID:       
# Copyright:    
# License:      
#----------------------------------------------------------------------------

"""
This is a generic application for simulation programs

Requirements:
    This application requires Python 2.x

Dependencies:
    wxPython    (download installer for your platform here: https://wxpython.org/Phoenix/snapshot-builds/)
"""

#MJP TODO:
#    - probably no longer need to disable mouse-over during fast forward
#    - visual display of frontier & growth locations(?)
#    - implement save
#        - as txt
#        - as image
#            - show grid in saved image?
#    - grip scrolling
#
#Previous dependencies no longer used:
#    pyPdf2      (install with 'pip install pypdf2')
#    Pillow      (install with 'pip install Pillow')
#    reportlab   (install with 'pip install reportlab')


#import wxversion
#wxversion.select('3.0-msw', optionsRequired=True)
#wxversion.select('3.0-Phoenix', optionsRequired=True)
import wx
import wx.richtext as rt
import os
import random
import ast
import sys
import thread
import threading
import wx.aui
from wx.lib.pubsub import pub
#from wx.lib.pubsub import Publisher
import SimUtils
import SimCanvas
from version import SimAppVersion
import wx.stc


class SimRTC(wx.stc.StyledTextCtrl):
    def __init__(self, parent, app, id=-1, style=wx.VSCROLL|wx.HSCROLL, pos=wx.DefaultPosition, size=wx.DefaultSize): 
        wx.stc.StyledTextCtrl.__init__(self,parent, id=-1, style=style, pos=pos, size=size)
        self.Bind(wx.EVT_CHAR, app.OnChar)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.MouseHandler)

    def MouseHandler(self, evt):
        # Comment out the following line to disable clicking into this window, however note that this disallows selection and copying of text
        evt.Skip()
        return

    def VisibleWrite(self, msg):  
        self.SetReadOnly(False)
        self.AppendText(msg)
        self.SetReadOnly(True)
        self.ScrollToEnd()
        return
    

#assertMode = wx.PYAPP_ASSERT_DIALOG

# IDs
SIMFW_ID_STOP            = 100
SIMFW_ID_FOR             = 101
SIMFW_ID_FF              = 102
SIMFW_ID_ZOOMIN          = 103
SIMFW_ID_ZOOMOUT         = 104
SIMFW_ID_ZOOM_MODE       = 105
SIMFW_ID_SELECT_MODE     = 106
SIMFW_ID_DELETE_IN       = 107
SIMFW_ID_DELETE_OUT      = 108
SIMFW_ID_UNDO            = 109
SIMFW_ID_DISABLEFRONTIER = 110
SIMFW_ID_ENABLEFRONTIER  = 111
SIMFW_ID_COPY            = 112
SIMFW_ID_PASTE           = 113
SIMFW_ID_FR              = 114
SIMFW_ID_REW             = 115
SIMFW_ID_DRAW_MODE       = 116


#---------------------------------------------------------------------------


class SimApp(wx.App):
    def __init__(self):
        SimUtils.__init__(self)
        wx.App.__init__(self, redirect=False)

    def OnInit(self, title="Simple self-assembly simulator application"):
        self.ctrlPanel = None
        self.messageWin = None
        self.showMessages = True
        self.messageQueue = []
        self.useAdvancedColorPicker = False
        self.nondetMode = SimUtils.SIMFW_NONDET_REPORT

        #wx.Log_SetActiveTarget(wx.LogStderr())
        #self.SetAssertMode(assertMode)

        self.frame = wx.Frame(None, -1, title, pos=(50,50), size=(400,400),
                         style=wx.DEFAULT_FRAME_STYLE, name=title)
        self.CreateStatusBar()
        self.frame._mgr = wx.aui.AuiManager(self.frame)

        self.CreateMenu()

        self.frame.Show(True)

        self.CreateCanvas()
        self.frame._mgr.AddPane(self.canvas, wx.CENTER, 'Simulation Window')

        self.AddMessageWindow()

        self.frame.Bind(wx.EVT_CLOSE, self.OnCloseFrame)
        self.frame.Bind(wx.EVT_CHAR, self.OnChar)
        
        self.CreateToolbar()
        
        pub.getDefaultPublisher().subscribe(self.DoOutputMessage, "outputMessage")
        pub.getDefaultPublisher().subscribe(self.DoWarningDialog, "warningDialog")
        pub.getDefaultPublisher().subscribe(self.DoErrorDialog, "errorDialog")

        # so set the frame to a good size for showing stuff
        self.frame.SetSize((1200, 800))
        self.frame.SetFocus()
        self.window = self.frame

        self.SetTopWindow(self.frame)

        #wx.Log_SetActiveTarget(wx.LogStderr())
        #wx.Log_SetTraceMask(wx.TraceMessages)
        self.frame.Maximize()
        
        # tell the manager to 'commit' all the changes just made
        self.frame._mgr.Update()
        
        #self.canvas.InitializeDrawing()
        #self.canvas.Refresh()
        self.canvas.DrawAssemblyUpdate("new")
        self.canvas.ScrollToCoords((0,0))

        return True
    
    def CreateStatusBar(self):
        self.frame.CreateStatusBar(3)
        self.frame.SetStatusWidths([150,200,-1])

    def CreateMenu(self):
        menuBar = wx.MenuBar()
        menu = wx.Menu()
        item1 = menu.Append(-1, "&Load\tCtrl-L", "Load system")
        self.Bind(wx.EVT_MENU, self.OnLoad, item1)
        itemR = menu.Append(-1, "Reload files", "Reload system from files")
        self.Bind(wx.EVT_MENU, self.OnReload, itemR)
        itemReset = menu.Append(-1, "&Reset to seed\tCtrl-R", "Reset system to seed")
        self.Bind(wx.EVT_MENU, self.OnReset, itemReset)
        item2 = menu.Append(-1, "&Save\tCtrl-S", "Save system")
        self.Bind(wx.EVT_MENU, self.OnSave, item2)
        menu.AppendSeparator()
        item3 = menu.Append(-1, "E&xit\tCtrl-X", "Exit simulator")
        self.Bind(wx.EVT_MENU, self.OnExitApp, item3)
        menuBar.Append(menu, "&File")

        menuCtrl = wx.Menu()
        self.radioIgnoreNondet = wx.MenuItem(menuCtrl, -1, text = "Ignore nondeterminism",kind = wx.ITEM_RADIO) 
        self.radioReportNondet = wx.MenuItem(menuCtrl, -1, text = "Report nondeterminism",kind = wx.ITEM_RADIO)
        self.radioStopAtNondet = wx.MenuItem(menuCtrl, -1, text = "Halt on nondeterminism",kind = wx.ITEM_RADIO)
        menuCtrl.Append(self.radioIgnoreNondet)
        menuCtrl.Append(self.radioReportNondet)
        menuCtrl.Append(self.radioStopAtNondet)
        self.radioReportNondet.Check()
        self.Bind(wx.EVT_MENU, self.OnNondetMode, self.radioIgnoreNondet)
        self.Bind(wx.EVT_MENU, self.OnNondetMode, self.radioReportNondet)
        self.Bind(wx.EVT_MENU, self.OnNondetMode, self.radioStopAtNondet)
        menuCtrl.AppendSeparator()
        self.checkKeepCache = wx.MenuItem(menuCtrl, -1, text = "Keep cache for rewinding", kind = wx.ITEM_CHECK)
        menuCtrl.Append(self.checkKeepCache)
        self.checkKeepCache.Check(True)
        self.Bind(wx.EVT_MENU, self.OnKeepCache, self.checkKeepCache)
        menuCtrl.AppendSeparator()
        itemFR = menuCtrl.Append(-1, "Fast rewind", "Fast rewind")
        self.Bind(wx.EVT_MENU, self.OnFastRewind, itemFR)
        itemStepRew = menuCtrl.Append(-1, "Rewind", "Rewind one step")
        self.Bind(wx.EVT_MENU, self.OnRewind, itemStepRew)
        itemStop = menuCtrl.Append(-1, "Pause", "Stop simulation")
        self.Bind(wx.EVT_MENU, self.OnQuit, itemStop)
        itemStep = menuCtrl.Append(-1, "&Forward\tCtrl-F", "Forward one step")
        self.Bind(wx.EVT_MENU, self.OnForward, itemStep)
        itemFF = menuCtrl.Append(-1, "&Go\tCtrl-G", "Fast forward")
        self.Bind(wx.EVT_MENU, self.OnGo, itemFF)
        itemQuit = menuCtrl.Append(-1, "&Quit\tCtrl-Q", "Quit fast forward")
        self.Bind(wx.EVT_MENU, self.OnQuit, itemQuit)
        menuBar.Append(menuCtrl, "&Control")

        self.menuView = wx.Menu()
        itemSTC = self.menuView.Append(-1, "&Center\tCtrl+C", "Scroll to center")
        self.Bind(wx.EVT_MENU, self.OnCenter, itemSTC)
        itemCOC = self.menuView.Append(-1, "&Center on coordinates...", "Scroll coordinates to center")
        self.Bind(wx.EVT_MENU, self.OnCenterOnCoords, itemCOC)
        self.menuView.AppendSeparator()
        itemZIn = self.menuView.Append(-1, "Zoom in\tCtrl-+", "Zoom in")
        self.Bind(wx.EVT_MENU, self.OnZoomIn, itemZIn)
        itemZOut = self.menuView.Append(-1, "Zoom out\tCtrl--", "Zoom out")
        self.Bind(wx.EVT_MENU, self.OnZoomOut, itemZOut)
        self.menuView.AppendSeparator()
        itemTG = self.menuView.Append(-1, "&Toggle grid lines\tCtrl-T", "Toggle display of grid lines")
        self.Bind(wx.EVT_MENU, self.OnToggleGrid, itemTG)
        itemTD = self.menuView.Append(-1, "&Toggle display updates\tCtrl-D", "Toggle display of tile additions")
        self.Bind(wx.EVT_MENU, self.OnToggleDisplay, itemTD)
        itemSTT = self.menuView.Append(-1, "Show only matching tile types...", "Set patterns for tile type names to display")
        self.Bind(wx.EVT_MENU, self.OnShowMatching, itemSTT)
        self.menuView.AppendSeparator()
        itemBGC = self.menuView.Append(-1, "Set background color", "Set background color")
        self.Bind(wx.EVT_MENU, self.OnSetBackgroundColor, itemBGC)
        self.menuView.AppendSeparator()
        itemRefresh = self.menuView.Append(-1, "Refresh screen\tF2", "Refresh screen")
        self.Bind(wx.EVT_MENU, self.OnRefresh, itemRefresh)

        menuBar.Append(self.menuView, "&View")

        menu2 = wx.Menu()
        itemAbout = menu2.Append(-1, "About", "About")
        self.Bind(wx.EVT_MENU, self.OnAbout, itemAbout)
        itemH = menu2.Append(-1, "&Help\tCtrl-H", "Show help")
        self.Bind(wx.EVT_MENU, self.OnHelp, itemH)
        menuBar.Append(menu2, "&Help")
        
        self.frame.SetMenuBar(menuBar)

    def AddMessageWindow(self):
        self.ctrlPanel = wx.Panel(self.frame, wx.ID_ANY, pos=wx.DefaultPosition, size=(300,200), style = wx.TAB_TRAVERSAL)
        self.messageWin = SimRTC(self.ctrlPanel, self, wx.ID_ANY, pos=wx.DefaultPosition, size=(500,200))#, style=wx.NOSCROLL|wx.HSCROLL|wx.NO_BORDER)
        #self.messageWin = rtcm.RichTextCtrl(self.ctrlPanel, id=wx.ID_ANY, style=wx.VSCROLL|wx.HSCROLL|rtcm.RE_READONLY, pos=wx.DefaultPosition, size=(500,70)) 
        #self.messageWin = rt.RichTextCtrl(self.ctrlPanel, wx.ID_ANY, pos=wx.DefaultPosition, size=(500,70))#, style=wx.NOSCROLL|wx.HSCROLL|wx.NO_BORDER)
        #self.messageWin.SetEditable(False)
        sz = wx.BoxSizer(wx.HORIZONTAL)
        sz.Add(self.messageWin, 20, wx.EXPAND, 20)
        self.ctrlPanel.SetSizerAndFit(sz)
        self.frame._mgr.Bind(wx.aui.EVT_AUI_PANE_CLOSE, self.OnCloseMessagePane)
        self.frame._mgr.AddPane(self.ctrlPanel, wx.BOTTOM, 'Message Panel')
        self.frame._mgr.Update()
 
    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("f") or keycode == ord("F"):
            self.canvas.DoForwardStep()
        elif keycode == ord("b") or keycode == ord("B"):
            self.canvas.DoRewindStep()
        elif keycode == ord("p") or keycode == ord("P"):
            self.canvas.PrintMessage()
        elif keycode == ord("r") or keycode == ord("R"):
            self.canvas.Reset(False)
        elif keycode == ord("s") or keycode == ord("S"):
            self.canvas.DoSave()
        elif keycode == ord("h") or keycode == ord("H"):
            self.canvas.ShowHelp()
        elif keycode == ord("l") or keycode == ord("L"):
            self.OnLoad(None)
        elif keycode == ord("=") or keycode == ord("+"):
            self.canvas.ZoomThenRecenter(True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.canvas.ZoomThenRecenter(False)
        elif keycode == ord("g") or keycode == ord("G"):
            self.canvas.DoGo()
        elif keycode == ord("q") or keycode == ord("Q"):
            self.canvas.DoQuit()
        elif keycode == ord("c") or keycode == ord("C"):
            self.canvas.ScrollToCoords((0,0))
        elif keycode == ord("t") or keycode == ord("T"):
            self.canvas.ToggleGrid()
        elif keycode == ord("d") or keycode == ord("D"):
            self.OnToggleDisplay(None)
        elif keycode == wx.WXK_F2:
            self.canvas.Refresh()
        elif keycode == wx.WXK_RETURN:
            self.canvas.DeleteSelection()
        elif keycode == wx.WXK_F2:
            self.canvas.Refresh()
        #elif keycode == wx.WXK_DELETE or keycode == wx.WXK_BACK:
            #self.DoDelete()

    def OnCloseMessagePane(self, event):
        if event.GetPane().window == self.ctrlPanel:
            self.menuView.Check(self.itemShowMess.GetId(), False)
        event.Skip()
        
    def OnShowMessageWin(self, event):
        pane = self.frame._mgr.GetPane(self.ctrlPanel)
        if self.itemShowMess.IsChecked():
            pane.Show()
        else:
            pane.Hide()
        self.frame._mgr.Update()

    def CreateCanvas(self):
        self.canvas = SimCanvas.Canvas(self.frame, wx.ID_ANY, self.frame, wx.DefaultSize, self)

    def CreateToolbar(self):
        # TODO: to make the spacing of the buttons look better, split this into multiple toolbars
        self.toolbar = wx.ToolBar(self.frame, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        self.toolbar.SetToolSeparation(20)
        self.toolbar.SetToolBitmapSize(wx.Size(24,24))
        self.toolbar.AddTool(SIMFW_ID_FR, "Fast Rewind", wx.Image("images/fastleft.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Fast Rewind")
        self.toolbar.AddTool(SIMFW_ID_REW, "Rewind", wx.Image("images/left.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Rewind")
        self.toolbar.AddTool(SIMFW_ID_STOP, "Pause", wx.Image("images/pause.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Pause")#, longHelp="Pause simulation")
        self.toolbar.AddTool(SIMFW_ID_FOR, "Forward", wx.Image("images/right.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Forward")#, longHelp="Forward one step")
        self.toolbar.AddTool(SIMFW_ID_FF, "Fast Forward", wx.Image("images/fastright.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Fast Forward")#longHelp="Fast forward")
        self.toolbar.AddSeparator()
        self.toolbar.AddTool(SIMFW_ID_ZOOMIN, "Zoom In", wx.Image("images/zoomin.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom In")#, longHelp="Zoom in")
        self.toolbar.AddTool(SIMFW_ID_ZOOMOUT, "Zoom Out", wx.Image("images/zoomout.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom Out")#, longHelp="Zoom out")
        self.toolbar.AddSeparator()
        btnImg = wx.Image("images/pan-zoom.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(45, 25)
        self.toolbar.AddTool(SIMFW_ID_ZOOM_MODE, "Zoom/Pan", btnImg.ConvertToBitmap(), shortHelp="Pan (left mouse)/zoom (right mouse)", kind=wx.ITEM_RADIO)
        btnImg = wx.Image("images/rectangle.xpm", wx.BITMAP_TYPE_XPM)
        btnImg.Rescale(25, 15)
        self.toolbar.AddTool(SIMFW_ID_SELECT_MODE, "Select Mode", btnImg.ConvertToBitmap(), shortHelp="Select Mode", kind=wx.ITEM_RADIO)
        btnImg = wx.Image("images/pencil.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(25, 15)
        self.toolbar.AddTool(SIMFW_ID_DRAW_MODE, "Draw Mode", btnImg.ConvertToBitmap(), shortHelp="Draw Mode", kind=wx.ITEM_RADIO)
        
        bmp = wx.Bitmap(20, 20)
        dc = wx.MemoryDC()
        dc.SelectObject(bmp)
        dc.Clear()
        color = wx.TheColourDatabase.Find("Red")
        brush = wx.Brush(color)
        dc.SetBrush(brush)
        dc.DrawCircle(10, 10, 10)
        dc.SelectObject(wx.NullBitmap)
        btnImg = bmp.ConvertToImage()
        self.toolbar.AddTool(SIMFW_ID_DISABLEFRONTIER, "Disable frontier locations", btnImg.ConvertToBitmap(), shortHelp="Disable frontier locations")
        self.toolbar.EnableTool(SIMFW_ID_DISABLEFRONTIER, False)
        
        bmp = wx.Bitmap(20, 20)
        dc = wx.MemoryDC()
        dc.SelectObject(bmp)
        dc.Clear()
        color = wx.TheColourDatabase.Find("Green")
        brush = wx.Brush(color)
        dc.SetBrush(brush)
        dc.DrawCircle(10, 10, 10)
        dc.SelectObject(wx.NullBitmap)
        btnImg = bmp.ConvertToImage()
        self.toolbar.AddTool(SIMFW_ID_ENABLEFRONTIER, "Enable frontier locations", btnImg.ConvertToBitmap(), shortHelp="Enable frontier locations")
        self.toolbar.EnableTool(SIMFW_ID_ENABLEFRONTIER, False)

        tsize = (20,20)
        btnImg = wx.ArtProvider.GetBitmap(wx.ART_COPY, wx.ART_TOOLBAR, tsize)
        self.toolbar.AddTool(SIMFW_ID_COPY, "Copy selection", btnImg, shortHelp="Copy selected tiles")
        self.toolbar.EnableTool(SIMFW_ID_COPY, False)
        btnImg = wx.ArtProvider.GetBitmap(wx.ART_PASTE, wx.ART_TOOLBAR, tsize)
        self.toolbar.AddTool(SIMFW_ID_PASTE, "Paste selection", btnImg, shortHelp="Paste selected tiles")
        self.toolbar.EnableTool(SIMFW_ID_PASTE, False)
        
        btnImg = wx.Image("images/delete-selection.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(20, 20)
        self.toolbar.AddTool(SIMFW_ID_DELETE_IN, "Delete inside selection", btnImg.ConvertToBitmap(), shortHelp="Delete selected tiles")
        self.toolbar.EnableTool(SIMFW_ID_DELETE_IN, False)
        btnImg = wx.Image("images/keep-selection.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(20, 20)
        self.toolbar.AddTool(SIMFW_ID_DELETE_OUT, "Delete outside selection", btnImg.ConvertToBitmap(), shortHelp="Delete unselected tiles")
        self.toolbar.EnableTool(SIMFW_ID_DELETE_OUT, False)
        self.toolbar.AddTool(SIMFW_ID_UNDO, "Undo last delete", wx.Image("images/undo.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Undo last delete")
        self.toolbar.EnableTool(SIMFW_ID_UNDO, False)
        self.toolbar.Realize()
        self.toolbar.Bind(wx.EVT_TOOL, self.OnTool)
        self.frame._mgr.AddPane(self.toolbar, wx.aui.AuiPaneInfo().
                          Name("Simulation toolbar").Caption("Simulation Toolbar").
                          ToolbarPane().Top())

    def OnTool(self, evt):
        if evt.GetId() == SIMFW_ID_STOP:
            self.canvas.DoQuit()
        elif evt.GetId() == SIMFW_ID_FR:
            self.canvas.DoGoBackward()
        elif evt.GetId() == SIMFW_ID_REW:
            self.canvas.DoRewindStep()
        elif evt.GetId() == SIMFW_ID_FOR:
            self.canvas.DoForwardStep()
        elif evt.GetId() == SIMFW_ID_FF:
            self.canvas.DoGo()
        elif evt.GetId() == SIMFW_ID_ZOOMIN:
            self.canvas.ZoomThenRecenter(True)
        elif evt.GetId() == SIMFW_ID_ZOOMOUT:
            self.canvas.ZoomThenRecenter(False)
        elif evt.GetId() == SIMFW_ID_ZOOM_MODE:
            self.SetMouseMode(SimCanvas.SIMFW_MODE_PAN_ZOOM)
        elif evt.GetId() == SIMFW_ID_SELECT_MODE:
            self.SetMouseMode(SimCanvas.SIMFW_MODE_SELECT)
        elif evt.GetId() == SIMFW_ID_DRAW_MODE:
            self.SetMouseMode(SimCanvas.SIMFW_MODE_DRAW)
        elif evt.GetId() == SIMFW_ID_COPY:
            self.canvas.CopySelection()
        elif evt.GetId() == SIMFW_ID_PASTE:
            self.canvas.PasteSelection()
        elif evt.GetId() == SIMFW_ID_ENABLEFRONTIER:
            self.canvas.EnableSelectedFrontier()
        elif evt.GetId() == SIMFW_ID_DISABLEFRONTIER:
            self.canvas.DisableSelectedFrontier()
        elif evt.GetId() == SIMFW_ID_DELETE_IN:
            self.canvas.DeleteSelection(bInside=True)
        elif evt.GetId() == SIMFW_ID_DELETE_OUT:
            self.canvas.DeleteSelection(bInside=False)
        elif evt.GetId() == SIMFW_ID_UNDO:
            self.UndoDelete()
            
    def SetMouseMode(self, mode):
        self.canvas.mouseMode = mode
        if mode == SimCanvas.SIMFW_MODE_PAN_ZOOM:
            self.canvas.rubberBand.reset()
            self.canvas.SetCursor(wx.Cursor(self.canvas.grabCursorImg))
            self.toolbar.EnableTool(SIMFW_ID_COPY, False)
            self.toolbar.EnableTool(SIMFW_ID_PASTE, False)
            self.toolbar.EnableTool(SIMFW_ID_DISABLEFRONTIER, False)
            self.toolbar.EnableTool(SIMFW_ID_ENABLEFRONTIER, False)
            self.toolbar.EnableTool(SIMFW_ID_DELETE_IN, False)
            self.toolbar.EnableTool(SIMFW_ID_DELETE_OUT, False)
            self.toolbar.EnableTool(SIMFW_ID_UNDO, False)
        else:
            self.canvas.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
            self.toolbar.EnableTool(SIMFW_ID_COPY, True)
            self.toolbar.EnableTool(SIMFW_ID_PASTE, True)
            self.toolbar.EnableTool(SIMFW_ID_DISABLEFRONTIER, True)
            self.toolbar.EnableTool(SIMFW_ID_ENABLEFRONTIER, True)
            self.toolbar.EnableTool(SIMFW_ID_DELETE_IN, True)
            self.toolbar.EnableTool(SIMFW_ID_DELETE_OUT, True)
            self.toolbar.EnableTool(SIMFW_ID_UNDO, True)
        
    def UndoDelete(self):
        for loc in self.canvas.lastDeletedList:
            self.canvas.assembly.frontier.discard(loc)
            self.canvas.assembly.AddTile(loc.type, loc.coords)
        self.canvas.lastDeletedList = []
        self.canvas.DoRedrawAssembly()

    def OutputMessage(self, msg):
        self.messageQueue.append(msg)
        if (self.messageWin is None) or (not self.showMessages):
            return
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "outputMessage")
        
    def DoOutputMessage(self):
        if not self.messageWin is None:
            if len(self.messageQueue) == 0:
                return
            allMsgs = ""
            for msg in self.messageQueue:
                if len(msg) > 0:
                    if not msg.endswith("\n"):
                        msg = msg + "\n"
                    allMsgs = allMsgs + msg
            self.messageQueue = []

            self.messageWin.VisibleWrite(allMsgs)
            #self.messageWin.ScrollIntoView(self.messageWin.GetLastPosition())

    def OnExitApp(self, evt):
        dlg = wx.MessageDialog(self.canvas, "Are you sure you want to exit?")
        if dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            self.frame.Close(True)
        dlg.Destroy()
        
    def OnNew(self, evt):
        self.canvas.DoNew()

    def OnLoad(self, evt):
        self.canvas.DoLoad()
        
    def OnLoadAssembly(self, evt):
        self.canvas.DoLoadAssembly()
        
    def OnReload(self, evt):
        self.canvas.Reset(True)
        
    def OnReset(self, evt):
        self.canvas.Reset(False)
        
    def OnSetAsSeed(self, event):
        self.canvas.DoSetAsSeed()

    def OnSave(self, event):
        self.canvas.DoSave()
        
    def OnSaveAs(self, event):
        self.canvas.DoSaveAs()
        
    def OnNondetMode(self, event):
        if self.radioIgnoreNondet.IsChecked():
            self.nondetMode = SimUtils.SIMFW_NONDET_IGNORE
            self.OutputMessage("Instances of nondeterminism during simulation will be ignored.")
        elif self.radioReportNondet.IsChecked():
            self.nondetMode = SimUtils.SIMFW_NONDET_REPORT
            self.OutputMessage("Instances of nondeterminism during simulation will be reported.")
        elif self.radioStopAtNondet.IsChecked():
            self.nondetMode = SimUtils.SIMFW_NONDET_STOP
            self.OutputMessage("Instances of nondeterminism during simulation will be reported and cause simulation to stop.")

    def OnKeepCache(self, event):
        self.SetCaching(self.checkKeepCache.IsChecked())
            
    def SetCaching(self, bOn):
        self.canvas.assembly.bKeepCache = bOn
        self.checkKeepCache.Check(bOn)
        self.toolbar.EnableTool(SIMFW_ID_FR, bOn)
        self.toolbar.EnableTool(SIMFW_ID_REW, bOn) 

    def OnFastRewind(self, event):
        self.canvas.DoGoBackward()

    def OnRewind(self, event):
        self.canvas.DoRewindStep()

    def OnForward(self, event):
        self.canvas.DoForwardStep()
        
    def OnGo(self, event):
        self.canvas.DoGo()
        
    def OnQuit(self, event):
        self.canvas.DoQuit()
            
    def OnToggleGrid(self, event):
        self.canvas.ToggleGrid()
        
    def OnToggleDisplay(self, event):
        self.canvas.ToggleDisplay()
        
    def OnShowMatching(self, event):
        self.canvas.SetShowMatching()
        
    def OnSetBackgroundColor(self, event):
        self.canvas.SetBackgroundColor()

    def OnRefresh(self, event):
        self.canvas.Refresh()
             
    def OnZoomIn(self, event):
        self.canvas.ZoomThenRecenter(True)

    def OnZoomOut(self, event):
        self.canvas.ZoomThenRecenter(False)
        
    def OnCenter(self, event):
        self.canvas.ScrollToCoords((0,0))
        
    def OnCenterOnCoords(self, event):
        doAgain = True
        coords = (0,0)
        while doAgain == True:
            doAgain = False
            dlg = wx.TextEntryDialog(self.frame, "Enter the coordinates to center on the screen\nin the form (x,y):", caption="Enter coordinates", value=str(coords))
            if not dlg.ShowModal() == wx.ID_OK:
                return
            result = dlg.GetValue()
            dlg.Destroy()
            try:
                c = eval(result)
                coords = (c[0],c[1])
            except:
                doAgain = True
        self.canvas.ScrollToCoords(coords)

    def OnAbout(self, event):
        versionString = SimAppVersion + "blah"
        aboutInfo = wx.adv.AboutDialogInfo()
        aboutInfo.SetName("Simulation App")
        aboutInfo.SetVersion(versionString)
        aboutInfo.SetDescription("wxPython-based simulation application")
        #aboutInfo.SetCopyright("(C) 2017-2019")
        aboutInfo.SetWebSite("http://www.self-assembly.net")
        aboutInfo.AddDeveloper("Matthew Patitz")
        wx.adv.AboutBox(aboutInfo)

    def OnHelp(self, event):
        self.canvas.ShowHelp()
        
    def ShowWarningDialog(self, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "warningDialog", msg=msg)
        
    def DoWarningDialog(self, msg):
        dlg = wx.MessageDialog(self.frame, msg, caption="Warning", style=wx.ICON_WARNING|wx.OK)
        dlg.ShowModal()
        dlg.Destroy()
        
    def ShowErrorDialog(self, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "errorDialog", msg=msg)

    def DoErrorDialog(self, msg):
        dlg = wx.MessageDialog(self.frame, msg, caption="Error", style=wx.ICON_ERROR|wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def OnCloseFrame(self, evt):
        # Make sure the simulation thread is killed
        self.canvas.DoQuit()
        timer = wx.Timer(self)
        timer.Start(1000)
            #self.canvas.goThread.exit()

        # deinitialize the frame manager
        self.frame._mgr.UnInit()
        # delete the frame
        self.frame.Destroy()

#----------------------------------------------------------------------------
# The following lines create the actual application and begin its execution
# Include them (uncommented!) in your child class of SimApp
#app = SimApp()
#app.MainLoop()

#---------------------------------------------------------------------------


